/*
 * bb_boxc.c : bearerbox box connection module
 *
 * handles start/restart/stop/suspend/die operations of the sms and
 * wapbox connections
 *
 * Kalle Marjola 2000 for project Kannel
 */

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <signal.h>

#include "gwlib/gwlib.h"
#include "msg.h"
#include "sms.h"
#include "bearerbox.h"
#include "wapbox.h"
#include "smsbox.h"
#include "bb_smscconn_cb.h"

/* passed from bearerbox core */

extern volatile sig_atomic_t bb_status;
extern List *incoming_sms;
extern List *outgoing_sms;
extern List *incoming_wdp;
extern List *outgoing_wdp;

extern List *flow_threads;
extern List *suspended;



/* our own thingies */

static volatile sig_atomic_t smsbox_running;
static volatile sig_atomic_t wapbox_running;
/* MJ - ASL only one wapbox and smsbox for a client */


/* MJ - Ports used in incoming SMS messages */
enum {
    CONNECTIONLESS_PUSH_CLIPORT = 2948,
    CONNECTIONLESS_SERVPORT = 9200,
    CONNECTED_CLIPORT = 9209,
    CONNECTED_SERVPORT = 9201
};

extern List *sms8bitreassembly;

/*------------------------------------------------
 * main single thread functions
 */

typedef struct _addrpar {
    Octstr *address;
    int port;
    int wapboxid;
} AddrPar;

static void ap_destroy(AddrPar *addr) {
    octstr_destroy(addr->address);
    gw_free(addr);
}

static int cmp_route(void *ap, void *ms) {
    AddrPar *addr = ap;
    Msg *msg = ms;

    if (msg->wdp_datagram.source_port == addr->port  &&
        octstr_compare(msg->wdp_datagram.source_address, addr->address)==0)
        return 1;

    return 0;
}


/*
 * this thread listens to incoming_wdp list
 * and then routs messages to proper wapbox
 */
static void wdp_to_wap(void *arg, List *wapIncoming) {
    AddrPar *ap;
    Msg *msg;
    int i;

    list_add_producer(flow_threads);

    while (bb_status != BB_DEAD) {

        list_consume(suspended);    /* block here if suspended */

        if ((msg = list_consume(incoming_wdp)) == NULL)
            break;

        gw_assert(msg_type(msg) == wdp_datagram);

        list_produce(wapIncoming, msg);
    }
    debug("bb", 0, "wdp_to_wapboxes: destroying lists");
    list_remove_producer(flow_threads);
}




/*-------------------------------------------------------------
 * public functions
 *
 * SMSBOX
 */

int smsbox_start(Cfg *cfg, List *smsOutgoing, List *smsIncoming) {
    CfgGroup *grp;
    SmsStartInfo *startinfo;

    if (smsbox_running) return -1;

    debug("bb", 0, "starting smsbox connection module");

    grp = cfg_get_single_group(cfg, octstr_imm("core"));

    list_add_producer(smsOutgoing);

    smsbox_running = 1;

    startinfo = gw_malloc(sizeof(SmsStartInfo));

    startinfo->cfg = cfg;
    startinfo->incoming_sms = smsIncoming;
    startinfo->outgoing_sms = smsOutgoing;


    if (gwthread_create(startsmsbox, (void *)startinfo) == -1)
        panic(0, "Failed to start a new thread for smsbox initialisation");


    return 0;
}





/* WAPBOX */

int wapbox_start(Cfg *cfg, List *wdpOutgoing, List *wdpIncoming) {
    CfgGroup *grp;
    WapStartInfo *startinfo;

    if (wapbox_running) return -1;

    debug("bb", 0, "starting wapbox connection module");

    grp = cfg_get_single_group(cfg, octstr_imm("core"));

    list_add_producer(wdpIncoming);

    startinfo = gw_malloc(sizeof(WapStartInfo));

    startinfo->cfg = cfg;
    startinfo->incoming_wdp = wdpIncoming;
    startinfo->outgoing_wdp = wdpOutgoing;

    if (gwthread_create(startwapbox, startinfo) == -1)
        panic(0, "Failed to start a new thread for wapbox initialisation");

    wapbox_running = 1;
    return 0;
}




/* MJ - ASL incoming SMS messages may be used in WAP context */

void sms8bit_destroy_item(void *item) {
    SMS8BitData *smsr = item;
    if (smsr->segments != NULL)
        list_destroy(smsr->segments, octstr_destroy_item);
    gw_free(smsr);
}

void parse_16bit_app_port_sms(Msg *msg, Octstr *informationelement) {
    msg->sms.destination_port = octstr_get_char(informationelement, 3);
    msg->sms.destination_port = msg->sms.destination_port + (256 * octstr_get_char(informationelement, 2));
    msg->sms.source_port = octstr_get_char(informationelement, 5);
    msg->sms.source_port = msg->sms.source_port + (256 * octstr_get_char(informationelement, 4));
}


static int find_8bitreassembly(void *struct_ptr, void *ref_ptr) {
    SMS8BitData *smsr = struct_ptr;
    int *refp = ref_ptr;
    return smsr->refnr == *refp;
}




Octstr *handle_8bitreference_reassembly(Octstr *data, Octstr *informationelement) {
    SMS8BitData *smsr;
    int refnr, seqnr, i;
    List *l;
    long ll;
    Octstr *reassembedMessage = NULL;

    refnr = octstr_get_char(informationelement, 2);
    seqnr = octstr_get_char(informationelement, 4);

    smsr = list_search(sms8bitreassembly, &refnr, find_8bitreassembly);
    if (smsr == NULL) {
        smsr = gw_malloc(sizeof(*smsr));
        smsr->refnr = refnr;
        smsr->curmsgs = 0;
        smsr->maxmsgs = octstr_get_char(informationelement, 3);
        smsr->segments = list_create();
        list_append(sms8bitreassembly, smsr);
    }

    ll = list_len(smsr->segments);
    if ((seqnr - 1) < ll)
        list_replace(smsr->segments, seqnr - 1, octstr_duplicate(data), NULL);
    else
        list_insert_expand(smsr->segments, seqnr - 1, octstr_duplicate(data));
    smsr->curmsgs++;

    if (smsr->curmsgs == smsr->maxmsgs) {
        /* complete */
        reassembedMessage = octstr_create("");
        for (i = 0; i < smsr->curmsgs; i++)
            octstr_append(reassembedMessage, (Octstr *)list_get(smsr->segments, i));

        /* clean up */

        /* TODO causes segfault */
        /* list_destroy(smsr->segments, octstr_destroy_item);
        list_delete_equal(sms8bitreassembly, smsr);
        gw_free(smsr); */
    }
    return reassembedMessage;
}



int parse_sms_headers(Msg *msg) {
    Octstr *ie;
    List *udhlist;
    int ie_identifier;
    Octstr *reassembedMsgData = NULL;
    int continuemessage = 1;

    /* list of information elements */
    udhlist = sms_get_user_data_headers(msg->sms.udhdata);

    /* see if this is */
    while ((ie = (Octstr *)list_extract_first(udhlist)) != NULL) {
        ie_identifier = octstr_get_char(ie, 0);
        switch (ie_identifier) {
            case IEI_16BITAPPLICATIONPORT:
                parse_16bit_app_port_sms(msg, ie);
                /* continue processing the message */
                break;
            case IEI_8BITCONCATENATED:
                /* add stuff for reassembling short messages */
                reassembedMsgData = handle_8bitreference_reassembly(msg->sms.msgdata, ie);
                if (reassembedMsgData != NULL) {
                    octstr_destroy(msg->sms.msgdata);
                    msg->sms.msgdata = reassembedMsgData;
                } else
                    /* don't do anything else with this message */
                    continuemessage = 0;
                break;
        }
    }
    /* TODO nother segfault 
    list_destroy(udhlist, octstr_destroy_item); */
    return continuemessage;
}


/*
 * Route the incoming message to a specific smsbox conn
 * or simply to a random if no shortcut routing and msg->sms.boxc_id
 * match.
 */
void route_incoming_sms(Msg *msg) {
    Octstr *s, *r;
    Msg *assembledmsg = NULL;

    s = r = NULL;
    gw_assert(msg_type(msg) == sms);

    /* msg_dump(msg, 0); */

    /* MJ - ASL check the type of SMS */
    switch (msg->sms.mclass) {
        case MC_CLASS1:
            /* check header data */
            if (msg->sms.udhdata != NULL) {
                if (parse_sms_headers(msg)) {
                    /* currently only handling connectionless push */
                    switch (msg->sms.destination_port) {
                        case CONNECTIONLESS_PUSH_CLIPORT:
                            /* connection wap push needs to be handled by wapbox */
                            list_produce(incoming_wdp, msg);
                            break;
                        default:
                            break;
                    }
                }
                else 
                    /* no longer needed */
                    msg_destroy(msg);
            }
            break;
        case MC_UNDEF:
            debug("bb_boxc.c", 0, "incoming sms");
            list_produce(incoming_sms, msg);
            break;
        default:
            debug("bb", 4, "Unsupported MMS message class: %d", msg->sms.mclass);
            break;
    }
}

